/**
 *  '$RCSfile: DataStreamTest.java,v $'
 *    Purpose: A test driver for the DataFileServer class
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Chad Berkley
 *    Release: @release@
 *
 *   '$Author: berkley $'
 *     '$Date: 2001/01/18 23:15:20 $'
 * '$Revision: 1.8 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.io.*;
import java.lang.*;
import java.util.*;
import java.sql.*;
import java.net.*;
import edu.ucsb.nceas.metacat.*;
import javax.servlet.http.*;


public class DataStreamTest
{
  DataStreamTest()
  {
  
  }
  
  /**
   * sends a file to the data file server socket
   */
  public static String SendFile(String filename, String host, int port, 
                                String cookie) 
  {
    Socket echoSocket = null;
    OutputStream out = null;
    InputStream in = null;
    DataOutputStream dsout = null;
    InputStreamReader isr = null;
    
		String retmsg = "";
    
    System.out.println("host: " + host + " port: " + port + " filename: " +
                       filename + " cookie: " + cookie);
    try 
    {
      echoSocket = DataFileServer.getSocket(host, port); //get the socket
      while(echoSocket == null) 
      {//loop until the port is there
        echoSocket = DataFileServer.getSocket(host, port);
      }
      //echoSocket = new Socket(host, port);
      out = echoSocket.getOutputStream(); //out to the server
      in = echoSocket.getInputStream();   //in from server
      isr = new InputStreamReader(in);
    } 
    catch (UnknownHostException e) 
    {
      System.err.println("Don't know about host: " + host + 
                         " : error in DataStreamTest.sendFile: " + 
                         e.getMessage());
      e.printStackTrace(System.out);
      System.exit(1);
    } 
    catch (IOException e) 
    {
      System.err.println("IO error in DataStreamTest.sendFile: "
                         + "broken connection to: " + host);
      System.out.println("error: " + e.getMessage());
      e.printStackTrace(System.out);
      System.exit(1);
    }
	  
    try
    {  
      File file = new File(filename); //get the file from the local system
      dsout = new DataOutputStream(out); 
      FileInputStream fs = new FileInputStream(file);
      // first convert the filename to a byte array
      String fn = file.getName();
      byte[] fname = fn.getBytes();
      // now write the string bytes followed by a '0'
      for (int i=0;i<fname.length;i++) 
      {
        dsout.write(fname[i]);    
      }
      dsout.write(0);  // marks end of name info
      
      //write the session id to the stream
      byte[] cook = cookie.getBytes();
      for(int i=0; i<cook.length; i++)
      {
        dsout.write(cook[i]);
      }
      dsout.write(0); //follow it with a 0
      
      //write the length of the file followed by a 0
      long flength = file.length();
      System.out.println("file sized (B): " + flength);
      String sflength = String.valueOf(flength); //note that this is changed
                                                 //to a string
      
      byte[] fl = sflength.getBytes();
      for(int i=0; i<fl.length; i++)
      {
        dsout.write(fl[i]);
      }
      dsout.write(0); //following 0
      
      //dsout.write((int)flength);
      //dsout.write(0);
      dsout.flush();
      
      BufferedReader rin = new BufferedReader(isr);

      // now send the file data
      byte[] buf = new byte[1024];
      int cnt = 0;
      int i = 0;
      while (cnt!=-1) 
      {
        cnt = fs.read(buf); //read the file
        System.out.println("i = "+ i +" Bytes read = " + cnt);
        if (cnt!=-1) 
        {
          dsout.write(buf, 0, cnt); //write the byte to the server
        }
        i++;
      }
      fs.close();
      dsout.flush();
	  }
	  catch (Exception w) 
    {
      System.out.println("error in DataStreamTest.sendFile: " + w.getMessage());
    }
   
    try
    { //get the server's response
      while(!isr.ready()) 
      {
        //System.out.println("not ready");
      }
      int msg = isr.read();
      retmsg += (char)msg;
      System.out.print("Message from server: ");
      while(msg != -1)
      {
        System.out.print((char)msg);
        System.out.flush();
        msg = isr.read(); //get the message a character at a time
        retmsg += (char)msg;
      }
      System.out.println();
      dsout.flush();
      dsout.close();
    }
    catch(Exception e)
    {
      System.out.println("error in DataStreamTest.sendFile: " + e.getMessage());
    }
	  return retmsg;
	}
  
  public static void main(String[] args)
  {
    System.out.println("Starting DataStreamTest");
    try
    {
      String temp = null;
      String servlet = "/berkley/servlet/metacat";
      String protocol = "http://";
      String host = "dev.nceas.ucsb.edu";
      String server = protocol + host + servlet;
      String filename = null;
      if(args.length == 0)
      {
        filename = "/home/berkley/xmltodb/test.dat";
      }
      else
      {
        filename = args[0];
      }
      //login url
      String u1str = server + "?action=login&username=higgins&password" +
                      "=neetar&qformat=xml";
      System.out.println("url 1: " + u1str);
      URL u1 = new URL(u1str);
      HttpMessage msg = new HttpMessage(u1);
      msg.sendPostMessage();
      String cookie = msg.getCookie();
      //get the port to send the data to
      System.out.println("url 2: " + server + "?action=getdataport");
      URL u2 = new URL(server + "?action=getdataport");
      HttpMessage msg2 = new HttpMessage(u2);
      InputStream in = msg2.sendPostMessage();
      InputStreamReader isr = new InputStreamReader(in);
      char c;
      int i = isr.read();
      String temp2 = null;
      while(i != -1)
      { //get the server response to the getdataport request
        c = (char)i;
        temp2 += c;
        i = isr.read();
      }
      int temp3 = temp2.indexOf("<", 32);
      temp2 = temp2.substring(32, temp3); //parse out the port
      //the port number taken from the xml encoding
      temp2 = temp2.trim();
      int port = (new Integer(temp2)).intValue();
      //while(portIsAvailable(port)) {int x = 1;}
      int index = cookie.indexOf("JSESSIONID=");
      index = cookie.indexOf("=", index);
      int index2 = cookie.indexOf(";", index);
      //get just the session id information from the cookie
      cookie = cookie.substring(index+1, index2);
      
      String servermsg = SendFile(filename, host, port, cookie);
      System.out.println("server says (end): " + servermsg);
    }
    catch(Exception e)
    {
      System.out.println("error in main: " + e.getMessage());
      e.printStackTrace(System.out);
    }
  }
}
