/**
 *  '$RCSfile: ForceReplicationHandler.java,v $'
 *    Purpose: A class to asyncronously force the replication of each server 
 *             that has an entry in the xml_replication table.  When run, 
 *             this thread communicates with each server in the list and 
 *             solicites a read of an updated or newly inserted document 
 *             with a certain docid.
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Chad Berkley
 *    Release: @release@
 *
 *   '$Author: berkley $'
 *     '$Date: 2001/01/18 23:15:21 $'
 * '$Revision: 1.6 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
package edu.ucsb.nceas.metacat;

import java.util.*;
import java.io.*;
import java.sql.*;
import java.net.*;
import java.lang.*;
import java.text.*;

/**
 * A class to asyncronously force the replication of each server 
 * that has an entry in the xml_replication table.  When run, 
 * this thread communicates with each server in the list and 
 * solicites a read of an updated or newly inserted document 
 * with a certain docid.
 */
public class ForceReplicationHandler implements Runnable
{
  private Thread btThread;
  private Connection conn;
  private MetaCatUtil util = new MetaCatUtil();
  private String docid;
  private String action;
  private boolean dbactionFlag = true;
  
  /**
   * @param docid the docid to force replicate
   * @param the action that is being performed on the document (either 
   *        INSERT or UPDATE)
   */
  public ForceReplicationHandler(String docid, String action)
  {
    this.docid = docid;
    this.action = action;
    if(this.action.equals(""))
    {
      dbactionFlag = false;
    }
    
    btThread = new Thread(this);
    btThread.setPriority(Thread.MIN_PRIORITY);
    btThread.start();
  }
  
  /**
   * Use this constructor when the action is implied.
   */
  public ForceReplicationHandler(String docid)
  {
    this.docid = docid;
    dbactionFlag = false;
    btThread = new Thread(this);
    btThread.setPriority(Thread.MIN_PRIORITY);
    btThread.start();
  }
  
  public void run()
  {
    //System.out.println("in ForceReplicationHandler Thread");
    try
    {
      URL comeAndGetIt;
      conn = util.openDBConnection();
      Enumeration keys = (ReplicationHandler.buildServerList(conn)).keys();
      //get the list of servers
      while(keys.hasMoreElements())
      {
        String server = (String)(keys.nextElement());
        if(dbactionFlag)
        {
          MetacatReplication.replLog("force replicating to " + server);
          comeAndGetIt = new URL("http://" + server + 
                                 "?action=forcereplicate&server=" + 
                                 util.getOption("server") + 
                                 util.getOption("replicationpath") +
                                 "&docid=" + docid + "&dbaction=" +
                                  action);
        }
        else
        {
          MetacatReplication.replLog("force replicating (default action) to )" + 
                                      server);
          comeAndGetIt = new URL("http://" + server + 
                                 "?action=forcereplicate&server=" + 
                                 util.getOption("server") + 
                                 util.getOption("replicationpath") +
                                 "&docid=" + docid);
        }
        MetaCatUtil.debugMessage("sending message: " + comeAndGetIt.toString());
        String message = MetacatReplication.getURLContent(comeAndGetIt);
        //send out the url.  message is a dummy variable as the target of
        //the URL never directly replies to the request.  this simply
        //invoces a read request from the server to this local machine.
      }
      conn.close();
    }
    catch(Exception e)
    {
      System.out.println("error in ForceReplicationHandler.run: " + 
                          e.getMessage());
      e.printStackTrace(System.out);
    }
    MetaCatUtil.debugMessage("exiting ForceReplicationHandler Thread");
  }
}
